{-------------------------------------------------------------------}
{ DSBITREE.INC                                                      }
{ Copyright (c) 1993, 1994  Julian M. Bucknall. All Rights Reserved }
{-------------------------------------------------------------------}

{=TreeRecursionStack(Queue)===========================================
Two descendants, one from a TStack, the other from a TQueue, for
unrolling recursion in a TBinTree. They pass on their errors, if any,
to the owning tree.
19Sep93 JMB
======================================================================}
type
{$IFNDEF UseTreeRecursion}
  TreeRecursionStack = object(TStack)
    Tree : PBinTree;
    constructor Init(aTree : PBinTree);
    procedure Error(Code : word); virtual;
  end;
{$ENDIF}
  TreeRecursionQueue = object(TQueue)
    Tree : PBinTree;
    constructor Init(aTree : PBinTree);
    procedure Error(Code : word); virtual;
  end;
{--------}
{$IFNDEF UseTreeRecursion}
constructor TreeRecursionStack.Init(aTree : PBinTree);
  begin
    if not inherited Init(false) then
      begin
        aTree^.Error(edsTreeStackError);
        Fail;
      end
    else
      Tree := aTree;
  end;
{--------}
procedure TreeRecursionStack.Error(Code : word);
  begin
    Tree^.Error(edsTreeStackError);
  end;
{$ENDIF}
{--------}
constructor TreeRecursionQueue.Init(aTree : PBinTree);
  begin
    if not inherited Init(false) then
      begin
        aTree^.Error(edsTreeQueueError);
        Fail;
      end
    else
      Tree := aTree;
  end;
{--------}
procedure TreeRecursionQueue.Error(Code : word);
  begin
    Tree^.Error(edsTreeQueueError);
  end;
{---------------------------------------------------------------------}

{Notes: the TTreeCursor is a pointer and a boolean wrapped in one. In
        TP6, BP7, and the TPWs, pointers allocated on the heap have at
        least a granularity of 4 bytes, ie their offset always has the
        lower 2 bits clear. We use bit 0 of the pointer as a left
        child, right child indicator (left = 0, right = 1). Thus the
        TTreeCursor is a pointer to the parent's node and an indicator
        to the relevant child.
        The parent link field of a node (the PKC) is a pointer and two
        booleans wrapped in one. The pointer is the parent's node as
        for TTreeCursors, bit 0 is the child (so a node always knows
        which child it is) and we use bit 1 of the pointer as a color
        bit for red-black trees (black = 0, red = 1). This by the way
        violates pure OOP design where ancestor aren't supposed to
        'know' about their descendants, but as I wrote the binary
        tree implementations in one go...
        The following 6 routines all help maintain these 'packed'
        variables.                                                    }

{-Given a cursor, returns the address of node's parent node}
function Dad(X : TTreeCursor) : PNode;
  inline($58/            {pop ax      get offset}
         $25/$FC/$FF/    {and ax, XX  clear color and child bits}
         $5A);           {pop dx      get seg/sel}
{--------}
{-Given a cursor, returns the child relationship the node has with its parent}
function Kid(X : TTreeCursor) : TChild;
  inline($58/            {pop ax      get offset}
         $25/$01/$00/    {and ax, 1   isolate child bit}
         $5A);           {pop dx      toss seg/sel}
{--------}
{-Given a cursor, returns the address of the node being pointed to}
function GetNode(Cursor : TTreeCursor) : PNode; near;
assembler;
  asm
    mov ax, Cursor.Word[2]
    mov es, ax
    mov di, Cursor.Word[0]
    mov ax, di
    and ax, $FFFC
    xchg ax, di
    and ax, 1
    shl ax, 1
    shl ax, 1
    add di, ax
    mov ax, es:[di+4]
    mov dx, es:[di+6]
  end;
{--------}
{-Converts a parent node and child relationship into a cursor}
function Csr(P : PNode; C : TChild) : TTreeCursor;
  inline($58/            {pop ax      get child}
         $25/$01/$00/    {and ax, 1   isolate child bit}
         $5B/            {pop bx      get offset}
         $09/$D8/        {or ax, bx   xfer child bit}
         $5A);           {pop dx      get seg/sel}
{--------}
{-Sets the cursor's color bit to zero}
function Bleach(Cursor : TTreeCursor) : TTreeCursor;
  inline ($58/           {pop ax      get offset}
          $25/$FD/$FF/   {and ax, XX  set off color bit}
          $5A);          {pop dx      get seg/sel}
{--------}
{-Sets the cursor's color bit to the same as a PKC link}
function Dye(Cursor, PKC : TTreeCursor) : TTreeCursor;
  inline ($58/             {pop ax      get color word}
          $25/$02/$00/     {and ax, 2   isolate color bit}
          $5B/             {pop bx      toss next}
          $5B/             {pop bx      get offset}
          $81/$E3/$FD/$FF/ {and bx, XX  kill color}
          $09/$D8/         {or ax, bx   xfer color bit}
          $5A);            {pop dx      get seg/sel}

{=TBinTree============================================================
A simple binary tree.

A binary tree is a data structure where each node has up to two
children, and one parent. This implementation makes a distinction
between external nodes (that have no children at all) and internal
nodes (that always have two children). External nodes are called
leaves. The object uses external cursors to navigate the tree (these
are NOT the nodes themselves). You position a given cursor in the tree
by moving it with the object's methods, and can use a cursor to insert
and delete data objects in the tree (although there are restrictions
on where this can happen).

The object has two iterators, and four methods to traverse the tree
with them. The four traversal methods are pre-order, in-order,
post-order and level-order. Note that JDS can be compiled in two modes
distinguished by the compiler define: UseTreeRecursion. If this is
active, recursive routines are used wherever required to implement
traversals; if not, then a TStack will be used to unravel the
recursion.

19Sep93 JMB
======================================================================}
{    Copyright (c) 1993, Julian M. Bucknall. All Rights Reserved      }
constructor TBinTree.Init(DataOwner : boolean);
  begin
    if not inherited Init(16, DataOwner) then Fail;

    TravType := InOrder;

    {These allocations always succeed}
    Rt := acNewNode(nil);
    NilNode := acNewNode(nil);

    Rt^.TLink[CLeft] := NilNode;
    Rt^.TLink[CRight] := NilNode;
    Rt^.PKC := 0;

    Count := 0;
  end;
{--------}
function  TBinTree.Clone(DataOwner : boolean;
                         NewCompare : TCompareFunc) : PAbstractContainer;
  var
    NewTree : PBinTree;
    NewNilNode : PNode;
    NoDataError : boolean;
  procedure DisposeSubTree(Node : PNode);
    begin
      if (Node <> NewNilNode) then
        begin
          DisposeSubTree(Node^.TLink[CLeft]);
          DisposeSubTree(Node^.TLink[CRight]);
          if DataOwner then
            NewTree^.DisposeData(Node^.Data);
          NewTree^.acDisposeNode(Node);
        end;
    end;
  function  MakeNewNode(Walker : TTreeCursor;
                        var NewNode : PNode;
                        NewParent : PNode) : boolean;
    var
      NewData   : pointer;
    begin
      if IsLeaf(Walker) then
        begin
          NewNode := NewNilNode;
          MakeNewNode := true;
        end
      else
        begin
          MakeNewNode := false;
          if DataOwner then
               NoDataError := NewTree^.DupData(Examine(Walker), NewData)
          else NewData := Examine(Walker);
          if not NoDataError then
            Exit;
          NewNode := NewTree^.acNewNode(NewData);
          if not Assigned(NewNode) then
            begin
              if DataOwner then
                NewTree^.DisposeData(NewData);
              NewNode := NewNilNode;
              Exit;
            end;
          NewNode^.PKC := (PNode(Walker)^.PKC and $00000003) or
                          longint(NewParent);
          if MakeNewNode(Left(Walker), NewNode^.TLink[CLeft], NewNode) then
            if MakeNewNode(Right(Walker), NewNode^.TLink[CRight], NewNode) then
              MakeNewNode := true
            else
              begin
                DisposeSubTree(NewNode);
                NewNode := NewNilNode;
              end;
        end;
    end;
  var
    JoinNode : PNode;
  begin
    NewTree := PBinTree(NewContainer(DataOwner));
    if Assigned(NewTree) then
      begin
        NewTree^.SetCompare(NewCompare);
        if (not IsEmpty) then
          begin
            NewNilNode := NewTree^.NilNode;
            NoDataError := true;
            if not MakeNewNode(Root, NewTree^.Rt^.TLink[CRight], NewTree^.Rt) then
              begin
                Dispose(NewTree, Done);
                NewTree := nil;
                Error(edsCloneError);
              end;
          end;
      end;
    Clone := NewTree;
  end;
{--------}
function TBinTree.Delete(Cursor : TTreeCursor) : TTreeCursor;
  var
    NewKid,
    LeftKid,
    RightKid : TTreeCursor;
    NodeToGo,
    Node : PNode;
  begin
    if IsLeaf(Cursor) then
      begin
        Delete := 0;
        Error(edsDelInvalidHere);
        Exit;
      end;
    RightKid := Right(Cursor);
    LeftKid := Left(Cursor);
    if not IsLeaf(RightKid) then
      if not IsLeaf(LeftKid) then
        begin
          Delete := 0;
          Error(edsDelInvalidHere);
          Exit;
        end
      else
        NewKid := RightKid
    else
      NewKid := LeftKid;
    Delete := Cursor;
    Node := GetNode(NewKid);
    NodeToGo := GetNode(Cursor);
    Dad(Cursor)^.TLink[Kid(Cursor)] := Node;
    if not IsLeaf(NewKid) then
      with Node^ do
        PKC := Dye(Cursor, PKC);
    acDisposeNode(NodeToGo);
  end;
{--------}
procedure TBinTree.Empty;
{$IFDEF UseTreeRecursion}
  {------}
  procedure RecursePostOrder(Cursor : TTreeCursor);
    begin
      if not IsLeaf(Cursor) then
        begin
          RecursePostOrder(Left(Cursor));
          RecursePostOrder(Right(Cursor));
          if OwnsData then
            DisposeData(Examine(Cursor));
          acDisposeNode(GetNode(Cursor));
        end;
    end;
  {------}
  begin
    RecursePostOrder(Root);
    Rt^.TLink[CRight] := NilNode;
    Count := 0;
    if InDone then
      begin
        acDisposeNode(Rt);
        acDisposeNode(NilNode);
      end;
  end;
{$ELSE}
  const
    Sentinel = 0;
  var
    Walker: TTreeCursor;
    Stack : TreeRecursionStack;
  begin
    if not IsEmpty then
      begin
        Stack.Init(@Self);
        Stack.Push(pointer(Root));
        repeat
          Walker := TTreeCursor(Stack.Examine);
          if (Walker = Sentinel) then
            begin
              Walker := TTreeCursor(Stack.Pop);
              Walker := TTreeCursor(Stack.Pop);
              if OwnsData then
                DisposeData(Examine(Walker));
              acDisposeNode(GetNode(Walker));
            end
          else
            begin
              Stack.Push(pointer(Sentinel));
              if not IsLeaf(Right(Walker)) then
                Stack.Push(pointer(Right(Walker)));
              if not IsLeaf(Left(Walker)) then
                Stack.Push(pointer(Left(Walker)));
            end;
        until (Stack.IsEmpty);
        Stack.Done;
        Rt^.TLink[CRight] := NilNode;
        Count := 0;
      end;
    if InDone then
      begin
        acDisposeNode(Rt);
        acDisposeNode(NilNode);
      end;
  end;
{$ENDIF}
{--------}
function TBinTree.Erase(Cursor : TTreeCursor) : TTreeCursor;
  begin
    if OwnsData then
      DisposeData(Examine(Cursor));
    Erase := Delete(Cursor);
  end;
{--------}
function  TBinTree.Examine(Cursor : TTreeCursor) : pointer;
  begin
    Examine := GetNode(Cursor)^.Data;
  end;
{--------}
function  TBinTree.FirstThat(Action : TIteratorFunc; ExtraData : pointer) : TTreeCursor;
  {------}
  function TraverseLevelOrder : TTreeCursor;
    var
      FoundIt : boolean;
      Walker: TTreeCursor;
      Queue : TreeRecursionQueue;
    begin
      TraverseLevelOrder := 0;
      FoundIt := false;
      Queue.Init(@Self);
      Queue.Append(pointer(Root));
      repeat
        Walker := TTreeCursor(Queue.Pop);
        if Action(@Self, Examine(Walker), ExtraData) then
          begin
            FoundIt := true;
            TraverseLevelOrder := Walker;
          end
        else
          begin
            if not IsLeaf(Left(Walker)) then
              Queue.Append(pointer(Left(Walker)));
            if not IsLeaf(Right(Walker)) then
              Queue.Append(pointer(Right(Walker)));
          end;
      until FoundIt or Queue.IsEmpty;
      Queue.Done;
    end;
  {------}
{$IFDEF UseTreeRecursion}
  function TraversePreOrder(Walker : TTreeCursor) : TTreeCursor;
    var
      Temp : TTreeCursor;
    begin
      if IsLeaf(Walker) then
        TraversePreOrder := 0
      else if Action(@Self, Examine(Walker), ExtraData) then
        TraversePreOrder := Walker
      else
        begin
          Temp := TraversePreOrder(Left(Walker));
          if (Temp <> 0) then
               TraversePreOrder := Temp
          else TraversePreOrder := TraversePreOrder(Right(Walker));
        end;
    end;
  {------}
  function TraverseInOrder(Walker : TTreeCursor) : TTreeCursor;
    var
      Temp : TTreeCursor;
    begin
      if IsLeaf(Walker) then
        TraverseInOrder := 0
      else
        begin
          Temp := TraverseInOrder(Left(Walker));
          if (Temp <> 0) then
            TraverseInOrder := Temp
          else if Action(@Self, Examine(Walker), ExtraData) then
            TraverseInOrder := Walker
          else
            TraverseInOrder := TraverseInOrder(Right(Walker));
        end;
    end;
  {------}
  function TraversePostOrder(Walker : TTreeCursor) : TTreeCursor;
    var
      Temp : TTreeCursor;
    begin
      if IsLeaf(Walker) then
        TraversePostOrder := 0
      else
        begin
          Temp := TraversePostOrder(Left(Walker));
          if (Temp <> 0) then
            TraversePostOrder := Temp
          else
            begin
              Temp := TraversePostOrder(Right(Walker));
              if (Temp <> 0) then
                TraversePostOrder := Temp
              else if Action(@Self, Examine(Walker), ExtraData) then
                TraversePostOrder := Walker
              else
                TraversePostOrder := 0;
            end;
        end;
    end;
  {------}
  begin
    case TravType of
      PreOrder   : FirstThat := TraversePreOrder(Root);
      InOrder    : FirstThat := TraverseInOrder(Root);
      PostOrder  : FirstThat := TraversePostOrder(Root);
      LevelOrder : FirstThat := TraverseLevelOrder;
    end;{case}
  end;
{$ELSE}
  const
    Sentinel = 0;
  function TraversePreOrder : TTreeCursor;
    var
      Walker: TTreeCursor;
      Stack : TreeRecursionStack;
      FoundIt : boolean;
    begin
      TraversePreOrder := 0;
      FoundIt := false;
      Stack.Init(@Self);
      Stack.Push(pointer(Root));
      repeat
        Walker := TTreeCursor(Stack.Pop);
        if Action(@Self, Examine(Walker), ExtraData) then
          begin
            TraversePreOrder := Walker;
            FoundIt := true;
          end
        else
          begin
            if not IsLeaf(Right(Walker)) then
              Stack.Push(pointer(Right(Walker)));
            if not IsLeaf(Left(Walker)) then
              Stack.Push(pointer(Left(Walker)));
          end;
      until FoundIt or Stack.IsEmpty;
      Stack.Done;
    end;
  {------}
  function TraverseInOrder : TTreeCursor;
    var
      Walker: TTreeCursor;
      Stack : TreeRecursionStack;
      FoundIt : boolean;
    begin
      TraverseInOrder := 0;
      FoundIt := false;
      Stack.Init(@Self);
      Stack.Push(pointer(Root));
      repeat
        Walker := TTreeCursor(Stack.Pop);
        if (Walker = Sentinel) then
          begin
            Walker := TTreeCursor(Stack.Pop);
            if Action(@Self, Examine(Walker), ExtraData) then
              begin
                TraverseInOrder := Walker;
                FoundIt := true;
              end;
          end
        else
          begin
            if not IsLeaf(Right(Walker)) then
              Stack.Push(pointer(Right(Walker)));
            Stack.Push(pointer(Walker));
            Stack.Push(pointer(Sentinel));
            if not IsLeaf(Left(Walker)) then
              Stack.Push(pointer(Left(Walker)));
          end;
      until FoundIt or Stack.IsEmpty;
      Stack.Done;
    end;
  {------}
  function TraversePostOrder : TTreeCursor;
    var
      Walker: TTreeCursor;
      Stack : TreeRecursionStack;
      FoundIt : boolean;
    begin
      TraversePostOrder := 0;
      FoundIt := false;
      Stack.Init(@Self);
      Stack.Push(pointer(Root));
      repeat
        Walker := TTreeCursor(Stack.Examine);
        if (Walker = Sentinel) then
          begin
            Walker := TTreeCursor(Stack.Pop);
            Walker := TTreeCursor(Stack.Pop);
            if Action(@Self, Examine(Walker), ExtraData) then
              begin
                TraversePostOrder := Walker;
                FoundIt := true;
              end;
          end
        else
          begin
            Stack.Push(pointer(Sentinel));
            if not IsLeaf(Right(Walker)) then
              Stack.Push(pointer(Right(Walker)));
            if not IsLeaf(Left(Walker)) then
              Stack.Push(pointer(Left(Walker)));
          end;
      until FoundIt or Stack.IsEmpty;
      Stack.Done;
    end;
  {------}
  begin
    if IsEmpty then
      FirstThat := 0
    else
      case TravType of
        PreOrder   : FirstThat := TraversePreOrder;
        InOrder    : FirstThat := TraverseInOrder;
        PostOrder  : FirstThat := TraversePostOrder;
        LevelOrder : FirstThat := TraverseLevelOrder;
      end;{case}
  end;
{$ENDIF}
{--------}
procedure TBinTree.ForEach(Action : TIteratorProc; ExtraData : pointer);
  {------}
  procedure TraverseLevelOrder;
    var
      Walker: TTreeCursor;
      Queue : TreeRecursionQueue;
    begin
      Queue.Init(@Self);
      Queue.Append(pointer(Root));
      repeat
        Walker := TTreeCursor(Queue.Pop);
        Action(@Self, Examine(Walker), ExtraData);
        if not IsLeaf(Left(Walker)) then
          Queue.Append(pointer(Left(Walker)));
        if not IsLeaf(Right(Walker)) then
          Queue.Append(pointer(Right(Walker)));
      until Queue.IsEmpty;
      Queue.Done;
    end;
  {------}
{$IFDEF UseTreeRecursion}
  procedure TraversePreOrder(Walker : TTreeCursor);
    begin
      if not IsLeaf(Walker) then
        begin
          Action(@Self, Examine(Walker), ExtraData);
          TraversePreOrder(Left(Walker));
          TraversePreOrder(Right(Walker));
        end;
    end;
  {------}
  procedure TraverseInOrder(Walker : TTreeCursor);
    begin
      if not IsLeaf(Walker) then
        begin
          TraverseInOrder(Left(Walker));
          Action(@Self, Examine(Walker), ExtraData);
          TraverseInOrder(Right(Walker));
        end;
    end;
  {------}
  procedure TraversePostOrder(Walker : TTreeCursor);
    begin
      if not IsLeaf(Walker) then
        begin
          TraversePostOrder(Left(Walker));
          TraversePostOrder(Right(Walker));
          Action(@Self, Examine(Walker), ExtraData);
        end;
    end;
  {------}
  begin
    case TravType of
      PreOrder   : TraversePreOrder(Root);
      InOrder    : TraverseInOrder(Root);
      PostOrder  : TraversePostOrder(Root);
      LevelOrder : TraverseLevelOrder;
    end;{case}
  end;
{$ELSE}
  const
    Sentinel = 0;
  {------}
  procedure TraversePreOrder;
    var
      Walker: TTreeCursor;
      Stack : TreeRecursionStack;
    begin
      Stack.Init(@Self);
      Stack.Push(pointer(Root));
      repeat
        Walker := TTreeCursor(Stack.Pop);
        Action(@Self, Examine(Walker), ExtraData);
        if not IsLeaf(Right(Walker)) then
          Stack.Push(pointer(Right(Walker)));
        if not IsLeaf(Left(Walker)) then
          Stack.Push(pointer(Left(Walker)));
      until Stack.IsEmpty;
      Stack.Done;
    end;
  {------}
  procedure TraverseInOrder;
    var
      Walker: TTreeCursor;
      Stack : TreeRecursionStack;
    begin
      Stack.Init(@Self);
      Stack.Push(pointer(Root));
      repeat
        Walker := TTreeCursor(Stack.Pop);
        if (Walker = Sentinel) then
          begin
            Walker := TTreeCursor(Stack.Pop);
            Action(@Self, Examine(Walker), ExtraData);
          end
        else
          begin
            if not IsLeaf(Right(Walker)) then
              Stack.Push(pointer(Right(Walker)));
            Stack.Push(pointer(Walker));
            Stack.Push(pointer(Sentinel));
            if not IsLeaf(Left(Walker)) then
              Stack.Push(pointer(Left(Walker)));
          end;
      until Stack.IsEmpty;
      Stack.Done;
    end;
  {------}
  procedure TraversePostOrder;
    var
      Walker: TTreeCursor;
      Stack : TreeRecursionStack;
    begin
      Stack.Init(@Self);
      Stack.Push(pointer(Root));
      repeat
        Walker := TTreeCursor(Stack.Examine);
        if (Walker = Sentinel) then
          begin
            Walker := TTreeCursor(Stack.Pop);
            Walker := TTreeCursor(Stack.Pop);
            Action(@Self, Examine(Walker), ExtraData);
          end
        else
          begin
            Stack.Push(pointer(Sentinel));
            if not IsLeaf(Right(Walker)) then
              Stack.Push(pointer(Right(Walker)));
            if not IsLeaf(Left(Walker)) then
              Stack.Push(pointer(Left(Walker)));
          end;
      until Stack.IsEmpty;
      Stack.Done;
    end;
  {------}
  begin
    if not IsEmpty then
      case TravType of
        PreOrder   : TraversePreOrder;
        InOrder    : TraverseInOrder;
        PostOrder  : TraversePostOrder;
        LevelOrder : TraverseLevelOrder;
      end;{case}
  end;
{$ENDIF}
{--------}
function TBinTree.Insert(var Cursor : TTreeCursor; aData : pointer) : boolean;
  var
    Node : PNode;
  begin
    Insert := false;
    if not IsLeaf(Cursor) then
      Error(edsInsInvalidHere)
    else
      begin
        Node := acNewNode(aData);
        if Assigned(Node) then
          with Node^ do
            begin
              PKC := Cursor;
              Dad(Cursor)^.TLink[Kid(Cursor)] := Node;
              Insert := true;
            end;
      end;
  end;
{--------}
function  TBinTree.IsLeaf(Cursor : TTreeCursor) : boolean;
  begin
    IsLeaf := GetNode(Cursor) = NilNode;
  end;
{--------}
function  TBinTree.IsRoot(Cursor : TTreeCursor) : boolean;
  begin
    IsRoot := Dad(Cursor) = Rt;
  end;
{--------}
procedure TBinTree.Join(Cursor : TTreeCursor; Tree : PBinTree);
{$IFDEF UseTreeRecursion}
  {------}
  procedure RecursePreOrder(TCursor : TTreeCursor);
    begin
      if Tree^.IsLeaf(TCursor) then
        Dad(TCursor)^.TLink[Kid(TCursor)] := NilNode
      else
        begin
          RecursePreOrder(Left(TCursor));
          RecursePreOrder(Right(TCursor));
        end;
    end;
  {------}
{$ELSE}
  var
    Walker: TTreeCursor;
    Stack : TreeRecursionStack;
{$ENDIF}
  var
    RootNode : PNode;
  begin
    if not IsLeaf(Cursor) then
      Error(edsInsInvalidHere)
    else
      if Assigned(Tree) then
        begin
          if not Tree^.IsEmpty then
            begin
              RootNode := GetNode(Tree^.Root);
              RootNode^.PKC := Cursor;
              Dad(Cursor)^.TLink[Kid(Cursor)] := RootNode;
              inc(Count, Tree^.Count);
              {patch up all the Tree^.NilNode's to ours}
              {$IFDEF UseTreeRecursion}
              RecursePreOrder(Cursor);
              {$ELSE}
              Stack.Init(@Self);
              Stack.Push(pointer(Cursor));
              repeat
                Walker := TTreeCursor(Stack.Pop);
                if Tree^.IsLeaf(Walker) then
                  Dad(Walker)^.TLink[Kid(Walker)] := NilNode
                else
                  begin
                    Stack.Push(pointer(Right(Walker)));
                    Stack.Push(pointer(Left(Walker)));
                  end;
              until (Stack.IsEmpty);
              Stack.Done;
              {$ENDIF}
              {patch up Tree}
              with Tree^ do
                begin
                  Rt^.TLink[CLeft] := NilNode;
                  Count := 0;
                end;
            end;
          Dispose(Tree, Done);
      end;
  end;
{--------}
function  TBinTree.Left(Cursor : TTreeCursor) : TTreeCursor;
  begin
    if IsLeaf(Cursor) then
      begin
        Left := 0;
        Error(edsCannotMoveHere);
      end
    else
      Left := Csr(GetNode(Cursor), CLeft);
  end;
{--------}
function  TBinTree.NewContainer(DataOwner : boolean) : PAbstractContainer;
  var
    NC : PAbstractContainer;
  begin
    NC := New(PBinTree, Init(DataOwner));
    if Assigned(NC) then
      begin
        NC^.SetCompare(Compare);
        NC^.SetDisposeData(DisposeData);
        NC^.SetDupData(DupData);
      end;
    NewContainer := NC;
  end;
{--------}
function  TBinTree.Parent(Cursor : TTreeCursor) : TTreeCursor;
  begin
    if IsRoot(Cursor) then
      begin
        Parent := 0;
        Error(edsCannotMoveHere);
      end
    else
      Parent := Bleach(Dad(Cursor)^.PKC);
  end;
{--------}
function  TBinTree.Replace(Cursor : TTreeCursor; aData : pointer) : pointer;
  begin
    if not IsLeaf(Cursor) then
       with GetNode(Cursor)^ do
         begin
           Replace := Data;
           Data := aData;
         end;
  end;
{--------}
function  TBinTree.Right(Cursor : TTreeCursor) : TTreeCursor;
  begin
    if IsLeaf(Cursor) then
      begin
        Right :=  0;
        Error(edsCannotMoveHere);
      end
    else
      Right := Csr(GetNode(Cursor), CRight);
  end;
{--------}
function  TBinTree.Root : TTreeCursor;
  begin
    Root := Csr(Rt, CRight);
  end;
{--------}
function  TBinTree.Search(var Cursor : TTreeCursor; aData : pointer) : boolean;
{$IFDEF UseTreeRecursion}
  {------}
  function RecursePreOrder(Walker : TTreeCursor) : boolean;
    begin
      if IsLeaf(Walker) then
        RecursePreOrder := false
      else if (Compare(Examine(Walker), aData) = 0) then
        begin
          RecursePreOrder := true;
          Cursor := Walker;
        end
      else if RecursePreOrder(Left(Walker)) then
        RecursePreOrder := true
      else
        RecursePreOrder := RecursePreOrder(Right(Walker));
    end;
  {------}
  begin
    Search := RecursePreOrder(Root);
  end;
{$ELSE}
  var
    Walker: TTreeCursor;
    Stack : TreeRecursionStack;
    FoundIt : boolean;
  begin
    FoundIt := false;
    Stack.Init(@Self);
    Stack.Push(pointer(Root));
    repeat
      Walker := TTreeCursor(Stack.Pop);
      if (Compare(Examine(Walker), aData) = 0) then
        begin
          FoundIt := true;
          Cursor := Walker;
        end
      else
        begin
          if not IsLeaf(Right(Walker)) then
            Stack.Push(pointer(Right(Walker)));
          if not IsLeaf(Left(Walker)) then
            Stack.Push(pointer(Left(Walker)));
        end;
    until FoundIt or Stack.IsEmpty;
    Stack.Done;
    Search := FoundIt;
  end;
{$ENDIF}
{--------}
procedure TBinTree.SetTraversalType(TT : TTraversalType);
  begin
    TravType := TT;
  end;
{--------}
function  TBinTree.acNewNode(aData : pointer) : PNode;
  var
    Node : PNode;
  begin
    Node := inherited acNewNode(aData);
    if Assigned(Node) then
      with Node^ do
        begin
           TLink[CLeft] := NilNode;
           TLink[CRight] := NilNode;
        end;
    acNewNode := Node;
  end;
{---------------------------------------------------------------------}
