program RedBlackVersusSkipList;
  {-Example program that tests speed and space performance of
    redblack trees versus skip lists. Also shows how to use
    these containers.

    Things to note especially:
      - the data dispose routine for the containers:
        MyDisposeData.
      - the use of the EZStrucs string compare routine:
        EZStrCompare.
      - in pmode, the tree can seem to have a memory leak.
        Untrue, the memory is allocated but not freed by RTM
        or DPMI16BI for their own uses.
      - the code that checks to see whether a new string was
        inserted, if not the string is freed.
      - the way the CountOfJs variable is passed to the
        CountStringsWithJ iterator routine through ForEach.}

uses
  EZRand, EZStrucs, Objects, DOS;

type
  PMyRBTree = ^TMyRBTree;
  TMyRBTree = object(TrbSearchTree)
    procedure Error(Code : word); virtual;
  end;

  PMySkipList = ^TMySkipList;
  TMySkipList = object(TSkipList)
    procedure Error(Code : word); virtual;
  end;

procedure TMyRBTree.Error(Code : word);
  {-Error method that ignores dup insertions.}
  begin
    if (Code <> edsInsertDup) then
      inherited Error(Code)
    else
      write('!');
  end;

procedure TMySkipList.Error(Code : word);
  {-Error method that ignores dup insertions.}
  begin
    if (Code <> edsInsertDup) then
      inherited Error(Code)
    else
      write('!');
  end;


function NewRandomStr : PString;
  {-Generate a new random string on the heap, 5<=Length<=25.}
  var
    i : integer;
    S : string[25];
  begin
    S[0] := char(jRandom(21) + 5);
    for i := 1 to length(S) do
      S[i] := char(jRandom(26) + ord('A'));
    NewRandomStr := NewStr(S);
  end;

procedure MyDisposeData(aData : pointer); far;
  {-Container data dispose procedure - disposes a PString}
  begin
    DisposeStr(PString(aData));
  end;

function GetTimeInSecs : longint;
  {-Return current time in seconds}
  var
    H, M, S, T : word;
  begin
    GetTime(H, M, S, T);
    GetTimeInSecs := (((longint(H)*60)+M)*60)+S + ord(T >= 50);
  end;

procedure CountStringsWithJ(C : PAbstractContainer;
                            aData : pointer;
                            ExtraData : pointer);
  {-Iterator to count the number of strings whose 2nd letter is 'J'}
  var
    S : PString absolute aData;
    CountOfJs : ^word absolute ExtraData;
  begin
    if (S^[2] = 'J') then
      inc(CountOfJs^);
  end;

function IsPalindrome(C : PAbstractContainer;
                      aData : pointer;
                      ExtraData : pointer) : boolean;
  {-Iterator to find the first palindromic string}
  var
    S : PString absolute aData;
    MidPoint : integer;
    i        : integer;
  begin
    MidPoint := length(S^) div 2;
    IsPalindrome := false;
    for i := 1 to MidPoint do
      if (S^[i] <> S^[length(S^) - i + 1]) then
        Exit;
    IsPalindrome := true;
  end;

const
  NumberOfStrings =
{$IFDEF MSDOS}
    5000;
{$ELSE}
    50000;
{$ENDIF}


var
  i : word;
  CountOfJs : word;

  S : PString;

  TreeCursor : TTreeCursor;
  ListCursor : TListCursor;

  MySkipList : PSkipList;
  MyRBTree   : PrbSearchTree;

  StartTime : longint;

  StartMem, MiddleMem : longint;

begin
  writeln('RED-BLACK TREE vs SKIP LIST test');

  writeln('Test1: redblack tree.');

  StartMem := MemAvail;
  MyRBTree := New(PMyRBTree, Init(true));
  with MyRBTree^ do
    begin
      SetCompare(EZStrCompare);
      SetDisposeData(MyDisposeData);
      SetTraversalType(InOrder);
    end;
  StartTime := GetTimeInSecs;

  writeln(' ..Insert');
  for i := 1 to NumberOfStrings do
    begin
      S := NewRandomStr;
      if not MyRBTree^.Insert(TreeCursor, S) then
        DisposeStr(S);
    end;
  writeln;

  MiddleMem := MemAvail;

  writeln(' ..ForEach');
  CountOfJs := 0;
  MyRBTree^.ForEach(CountStringsWithJ, @CountOfJs);
  writeln(' number of strings with 2nd char J: ', CountOfJs);

  writeln(' ..FirstThat');
  TreeCursor := MyRBTree^.FirstThat(IsPalindrome, nil);
  if (TreeCursor = 0) then
    writeln(' there are no palindromes')
  else
    writeln(' the first palindrome is: ',
            PString(MyRBTree^.Examine(TreeCursor))^);

  writeln(' ..Empty');
  MyRBTree^.Empty;

  writeln('  ..test time: ', GetTimeInSecs - StartTime, ' secs');
  Dispose(MyRBTree, Done);
  writeln(' ..memory used: ', StartMem - MiddleMem);
  if (MemAvail <> StartMem) then
    begin
      writeln('***Some memory not freed***');
      writeln('  (', StartMem - MemAvail, ' bytes)');
    end;

  writeln;
  writeln('Test2: skiplist.');

  StartMem := MemAvail;
  MySkipList := New(PMySkipList, Init(true));
  with MySkipList^ do
    begin
      SetCompare(EZStrCompare);
      SetDisposeData(MyDisposeData);
    end;
  StartTime := GetTimeInSecs;

  writeln(' ..Insert');
  for i := 1 to NumberOfStrings do
    begin
      S := NewRandomStr;
      if not MySkipList^.Insert(TreeCursor, S) then
        DisposeStr(S);
    end;
  writeln;

  MiddleMem := MemAvail;

  writeln(' ..ForEach');
  CountOfJs := 0;
  MySkipList^.ForEach(CountStringsWithJ, @CountOfJs);
  writeln(' number of strings with 2nd char J: ', CountOfJs);

  writeln(' ..FirstThat');
  ListCursor := MySkipList^.FirstThat(IsPalindrome, nil);
  if (ListCursor = 0) then
    writeln(' there are no palindromes')
  else
    writeln(' the first palindrome is: ',
            PString(MySkipList^.Examine(ListCursor))^);

  writeln(' ..Empty');
  MySkipList^.Empty;

  writeln('  ..test time: ', GetTimeInSecs - StartTime, ' secs');
  Dispose(MySkipList, Done);
  writeln(' ..memory used: ', StartMem - MiddleMem);
  if (MemAvail <> StartMem) then
    begin
      writeln('***Some memory not freed***');
      writeln('  (', StartMem - MemAvail, ' bytes)');
    end;
end.